from django.conf import settings
from django.db import models
from django.utils import timezone

from .enums import Channel


class Log(models.Model):
    created_at = models.DateTimeField(null=True, editable=False)
    checked_at = models.DateTimeField(null=True, editable=False)
    expires_at = models.DateTimeField(null=True, editable=False)
    token_type = models.CharField(max_length=255, null=True, editable=False)
    uidb64 = models.CharField(max_length=255, null=True, editable=False)
    hash = models.CharField(max_length=255, unique=True, null=True, editable=False)
    name = models.CharField(max_length=255, null=True, editable=False)
    email = models.EmailField(null=True, editable=False)
    phone = models.CharField(max_length=255, null=True, editable=False)
    channel = models.CharField(
        max_length=255, choices=Channel.choices, null=True, editable=False
    )
    send_precondition_failed = models.CharField(max_length=255, null=True, editable=False)
    check_precondition_failed = models.CharField(
        max_length=255, null=True, editable=False
    )
    sent = models.BooleanField(default=False, editable=False)
    errors = models.CharField(max_length=255, null=True, editable=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, editable=False
    )

    @property
    def checked(self) -> bool:
        return self.checked_at is not None

    def _check(self):
        if not self.checked:
            self.checked_at = timezone.now()
            self.save(update_fields=["checked_at"])



import base64
import json
import mimetypes
from os import getenv
from typing import Any

try:
    from sendgrid import SendGridAPIClient

    HAS_SENDGRID = True
except ImportError:
    HAS_SENDGRID = False

SENDGRID_API_KEY = getenv("SENDGRID_API_KEY") if HAS_SENDGRID else None


class SendgridAPI:
    def __init__(self, sender_name: str, sender_email: str):
        self.sender_name = sender_name
        self.sender_email = sender_email
        self._client = SendGridAPIClient(SENDGRID_API_KEY) if SENDGRID_API_KEY else None

    @staticmethod
    def _validate_personalizations(personalizations: list):
        # Remove duplicates
        personalizations = list({json.dumps(p) for p in personalizations})
        personalizations = [json.loads(p) for p in personalizations]

        return personalizations

    def send_mail(
        self,
        personalizations: list[dict[str, Any]],
        template_id: str,
        files: list | None = None,
        fail_silently: bool = False,
    ) -> int:
        personalizations = self._validate_personalizations(personalizations)

        # If no personalizations, return default response
        if not personalizations:
            return 0

        # Message payload
        message = {
            "personalizations": personalizations,
            "from": {
                "email": self.sender_email,
                "name": self.sender_name,
            },
            "template_id": template_id,
        }

        # Attachments
        attachments = []
        for file in files or []:
            try:
                data = file.read()
                encoded = base64.b64encode(data).decode()
            except (AttributeError, TypeError) as e:
                if fail_silently:
                    return 0

                raise e

            attachment = {
                "content": str(encoded),
                "type": mimetypes.guess_type(file.name)[0] or "application/octet-stream",
                "filename": str(file.name),
            }

            attachments.append(attachment)

        if attachments:
            message["attachments"] = attachments

        # Send
        try:
            response = self._client.send(message)
        except Exception as e:
            if fail_silently:
                return 0

            raise e

        if response.status_code == 202:
            return sum([len(p["to"]) for p in personalizations])

        return 0



import string
from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass, field
from datetime import datetime
from relattrs import rgetattr, rhasattr
from typing import Any

from jinja2 import Template as JinjaTemplate

from django.conf import settings
from django.db.utils import ProgrammingError
from django.utils import timezone
from django.utils.encoding import force_bytes, force_str
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
from django.utils.module_loading import import_string

from .enums import Channel
from .exceptions import URLTokenizerError
from .models import Log

SETTINGS = getattr(settings, "URL_TOKENIZER_SETTINGS", {})


def _from_config(config: dict[str, Any], key: str, default: Any) -> Any:
    return config.get(key, SETTINGS.get(key.upper(), default))


def _parse_preconditions(
    config: dict[str, Any], key: str
) -> dict[str, Callable[[object], bool]]:
    preconditions = SETTINGS.get(key.upper(), {})
    preconditions.update(SETTINGS.get("PRECONDITIONS", {}))
    preconditions.update(config.get(key, {}))
    preconditions.update(config.get("preconditions", {}))

    return {
        key: import_string(pred) if isinstance(pred, str) else pred
        for key, pred in preconditions.items()
    }


def parse_path(path: str | Callable[[object], str] | None, user: object) -> str:
    if callable(path):
        path = path(user)

    return (path or "").strip("/")


def encode(s: Any) -> str:
    return urlsafe_base64_encode(force_bytes(s))


def decode(s: bytes | str) -> str:
    return force_str(urlsafe_base64_decode(s))


@dataclass
class URLToken:
    user: object
    type: str
    created_at: datetime = field(default_factory=timezone.now)
    expires_at: datetime | None = None
    uidb64: str = ""
    token: str = ""
    link: str = ""
    hash: str | None = None
    email: str = ""
    name: str = ""
    phone: str = ""
    channel: Channel | None = None
    precondition_failed: str | None = None
    sent: bool = False
    exception: URLTokenizerError | None = None
    log: Log | None = None
    extra_tokens: dict = field(default_factory=dict)

    def _(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

        return self

    def _log(self) -> Log | None:
        with suppress(ProgrammingError):
            self.log = Log.objects.create(
                created_at=self.created_at,
                expires_at=self.expires_at,
                token_type=self.type,
                uidb64=self.uidb64,
                hash=self.hash,
                email=self.email,
                name=self.name,
                phone=self.phone,
                channel=self.channel,
                send_precondition_failed=self.precondition_failed,
                sent=self.sent,
                errors=self.exception.__repr__() if self.exception else None,
                user=self.user,
            )

            return self.log

        return None


class Template:
    def __init__(self, id: str, plain_content: str, context: dict[str, Any]):
        self.id = id
        self.plain_content = plain_content
        self.context = context
        self.params = self._params_from_plain_content

    @property
    def _params_from_plain_content(self) -> list[str]:
        def is_underscore(s: str) -> bool:
            return not s[0].isdigit() and all(
                c in string.ascii_lowercase + "_" + string.digits for c in s
            )

        params = set()
        for param in self.plain_content.split("{{")[1:]:
            param = param.split("}}")[0].replace(" ", "")
            if is_underscore(param) or param.strip().startswith("#each"):
                params.add(param.replace("#each", ""))

        return list(params)

    @staticmethod
    def _parse_context(
        url_token: URLToken,
        data: dict[str, Any] | Callable[[URLToken], dict[str, Any]] | None,
    ) -> dict[str, Any]:
        if data is None:
            return {}

        if callable(data):
            return data(url_token)

        return {
            key: value(url_token) if callable(value) else value
            for key, value in data.items()
        }

    def get_template_data(self, url_token: URLToken) -> dict[str, Any]:
        data = {
            k: rgetattr(url_token.user, k)
            for k in self.params
            if rhasattr(url_token.user, k)
        }
        data.update(self._parse_context(url_token, self.context))
        return data

    def render(self, url_token: URLToken) -> str:
        template = JinjaTemplate(
            self.plain_content.replace("{{{", "{{").replace("}}}", "}}")
        )
        return template.render(self.get_template_data(url_token))



import hashlib
import threading
from collections.abc import Iterable
from datetime import timedelta
from enum import Enum
from typing import Any, Callable

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.core.mail import send_mail
from django.utils import timezone
from django.utils.encoding import DjangoUnicodeDecodeError, force_bytes
from django.utils.translation import gettext_lazy as _

from .enums import Channel
from .exceptions import ErrorCode, URLTokenizerError
from .models import Log
from .sendgrid.api import SendgridAPI
from .token_generator import TokenGenerator
from .utils import (
    SETTINGS,
    Template,
    URLToken,
    _from_config,
    _parse_preconditions,
    decode,
    encode,
    parse_path,
)

try:
    from sms import send_sms

    HAS_SMS = True
except ImportError:
    HAS_SMS = False


class URLTokenizer:
    @property
    def user_model(self) -> object:
        return get_user_model()

    @property
    def encoding_field(self) -> str:
        return self._token_generator.encoding_field

    def __init__(self, token_type: str | Enum | None = None):
        self.token_type = self._parse_token_type(token_type)
        self.validate_token_type = SETTINGS.get("VALIDATE_TOKEN_TYPE", True)
        # at this point token_type is either None or a string

        # token
        token_config = self._get_token_config(SETTINGS, self.token_type)
        self._token_generator = TokenGenerator(token_config)
        # extra token types
        extra_token_types = _from_config(token_config, "extra_token_types", [])
        extra_token_types = [self._parse_token_type(t) for t in extra_token_types]
        extra_token_configs = {
            t: self._get_token_config(SETTINGS, t) for t in extra_token_types
        }
        self._extra_token_generators = {
            t: TokenGenerator(extra_token_configs[t]) for t in extra_token_types
        }

        # url
        self.path = _from_config(token_config, "path", "")
        self.domain = _from_config(token_config, "domain", "localhost")
        self.protocol = _from_config(token_config, "protocol", "http")
        self.port = _from_config(token_config, "port", "80")

        # sending
        self.send_enabled = _from_config(token_config, "send_enabled", False)
        self.channel = _from_config(token_config, "channel", None)
        self.send_preconditions = _parse_preconditions(token_config, "send_preconditions")

        # template
        template_id = _from_config(token_config, "template_id", None)
        plain_content = _from_config(token_config, "plain_content", "")
        template_data = _from_config(token_config, "template_data", {})
        self.template = Template(template_id, plain_content, template_data)

        # email
        self.email_field = _from_config(token_config, "email_field", "email")
        self.name_field = _from_config(token_config, "name_field", "name")
        self.email_subject = _from_config(
            token_config,
            "email_subject",
            "link generated with django-url-tokenizer",
        )

        # sendgrid
        self._sendgrid_api = SendgridAPI(
            _from_config(token_config, "sender_name", None), settings.DEFAULT_FROM_EMAIL
        )

        # sms
        self.phone_field = _from_config(token_config, "phone_field", "phone")

        # logging
        self.logging_enabled = _from_config(token_config, "logging_enabled", False)

        # error handling
        self.fail_silently_on_generate = _from_config(
            token_config, "fail_silently_on_generate", False
        )
        self.fail_silently_on_bulk_generate = _from_config(
            token_config, "fail_silently_on_bulk_generate", False
        )
        self.fail_silently_on_check = _from_config(
            token_config, "fail_silently_on_check", False
        )
        self.fail_silently_on_callbacks = _from_config(
            token_config, "fail_silently_on_callbacks", False
        )

    # initialization

    @staticmethod
    def _parse_token_type(token_type: str | Enum | None) -> str | None:
        if isinstance(token_type, str):
            token_type = token_type.strip().lower()
        elif isinstance(token_type, Enum):
            token_type = token_type.value.strip().lower()
        elif token_type is not None:
            raise ValueError(_("'token_type' must be either a string or Enum"))

        return token_type

    def _get_token_config(
        self, settings_: dict[str, Any], token_type: str | None
    ) -> dict[str, Any]:
        TOKEN_CONFIG = settings_.get("TOKEN_CONFIG", {})

        # avoid empty token_type
        if any(key.strip() == "" for key in TOKEN_CONFIG.keys()):
            raise ImproperlyConfigured(
                _("TOKEN_CONFIG cannot contain blank 'token_type'.")
            )

        if token_type is None:
            return TOKEN_CONFIG.get("default", {})

        # validate token_type
        token_config = TOKEN_CONFIG.get(token_type, None)
        if token_config is None and self.validate_token_type:
            raise URLTokenizerError(ErrorCode.invalid_token_type, token_type=token_type)

        return token_config or TOKEN_CONFIG.get("default", {})

    # helpers

    def _validate_preconditions(
        self, url_token: URLToken, fail_silently: bool = False
    ) -> bool:
        for k, pred in self.send_preconditions.items():
            try:
                if pred(url_token.user):
                    continue
            except Exception as e:
                url_token.exception = URLTokenizerError(
                    ErrorCode.send_precondition_execution_error,
                    context={"exception": e},
                    pred=k,
                )

            url_token.precondition_failed = k

            if self.logging_enabled:
                url_token._log()

            if url_token.exception and not fail_silently:
                from_exc = url_token.exception.context.get("exception")
                raise url_token.exception from from_exc

            return False

        return True

    def _send_link(
        self,
        url_token: URLToken,
        template: Template | Callable[[URLToken], Template] | None = None,
        email_subject: str | None = None,
        fail_silently: bool = False,
    ) -> URLToken:
        if callable(template) and not isinstance(template, Template):
            template = template(url_token)

        message = template.render(url_token) if template else None

        if url_token.channel == Channel.EMAIL:
            if not url_token.email:
                return url_token._(exception=URLTokenizerError(ErrorCode.no_email))

            # sendgrid
            if template and template.id and self._sendgrid_api._client:
                personalizations = [
                    {
                        "to": [{"email": url_token.email, "name": url_token.name}],
                        "dynamic_template_data": template.get_template_data(url_token),
                    }
                ]

                sent = self._sendgrid_api.send_mail(
                    personalizations,
                    template_id=template.id,
                    fail_silently=fail_silently,
                )

            # django send_mail
            else:
                sent = send_mail(
                    subject=email_subject,
                    message=message or url_token.link,
                    from_email=settings.DEFAULT_FROM_EMAIL,
                    recipient_list=[url_token.email],
                    fail_silently=fail_silently,
                )

            url_token.sent = sent > 0

        elif url_token.channel == Channel.SMS and HAS_SMS:
            if not url_token.phone:
                return url_token._(exception=URLTokenizerError(ErrorCode.no_phone))

            sent = send_sms(
                body=message or url_token.link,
                originator=settings.DEFAULT_FROM_SMS,
                recipients=[url_token.phone],
                fail_silently=fail_silently,
            )
            url_token.sent = sent > 0

        return url_token

    # main methods

    def generate_tokenized_link(
        self,
        user: object,
        path: str | Callable[[object], str] | None = None,
        domain: str | None = None,
        protocol: str | None = None,
        port: str | None = None,
        channel: Channel | None = None,
        template: Template | Callable[[URLToken], Template] | None = None,
        email_subject: str | None = None,
        fail_silently: bool | None = None,
    ) -> URLToken:
        path = parse_path(path or self.path, user)
        domain = domain or self.domain
        protocol = protocol or self.protocol
        port = port or self.port
        channel = channel or self.channel
        template = template or self.template
        email_subject = email_subject or self.email_subject
        if fail_silently is None:
            fail_silently = self.fail_silently_on_generate

        email = str(getattr(user, self.email_field, "") or "")
        name = str(getattr(user, self.name_field, "") or "")
        phone = str(getattr(user, self.phone_field, "") or "")
        url_token = URLToken(
            user, self.token_type, email=email, name=name, phone=phone, channel=channel
        )

        if not self._validate_preconditions(url_token, fail_silently):
            return url_token

        uidb64 = encode(getattr(user, self.encoding_field))
        token, ts = self._token_generator.make_token(user)
        link = f"{protocol}://{domain}:{port}/{path}?uid={uidb64}&key={token}"
        token_hash = hashlib.sha256(force_bytes(uidb64 + token)).hexdigest()
        expires_at = timezone.make_aware(ts) + timedelta(
            seconds=self._token_generator.timeout
        )

        for i, (extra_token_type, extra_token_generator) in enumerate(
            self._extra_token_generators.items(), start=2
        ):
            extra_token, _ = extra_token_generator.make_token(user)
            url_token.extra_tokens[extra_token_type] = extra_token
            link += f"&key{i}={extra_token}"

        url_token = url_token._(
            uidb64=uidb64,
            token=token,
            link=link,
            hash=token_hash,
            expires_at=expires_at,
        )

        if self.send_enabled:
            url_token = self._send_link(
                url_token,
                template=template,
                email_subject=email_subject,
                fail_silently=fail_silently,
            )

        if self.logging_enabled:
            url_token._log()

        if url_token.exception and not fail_silently:
            raise url_token.exception

        return url_token

    def bulk_generate_tokenized_link(
        self,
        users: Iterable[object],
        path: str | Callable[[object], str] | None = None,
        domain: str | None = None,
        protocol: str | None = None,
        port: str | None = None,
        channel: Channel | None = None,
        template: Template | Callable[[URLToken], Template] | None = None,
        email_subject: str | None = None,
        fail_silently: bool | None = None,
    ) -> list[URLToken]:
        url_tokens, threads = [], []

        if users is None:
            return url_tokens

        if fail_silently is None:
            fail_silently = self.fail_silently_on_bulk_generate

        # Define a helper function to execute generate_tokenized_link for each user
        def generate_link(user):
            url_token = self.generate_tokenized_link(
                user,
                path=path,
                domain=domain,
                protocol=protocol,
                port=port,
                channel=channel,
                template=template,
                email_subject=email_subject,
                fail_silently=fail_silently,
            )
            url_tokens.append(url_token)

        # Create a thread for each user
        for user in users:
            thread = threading.Thread(target=generate_link, args=(user,))
            threads.append(thread)

        # Start all the threads
        for thread in threads:
            thread.start()

        # Wait for all the threads to finish
        for thread in threads:
            thread.join()

        return url_tokens

    def check_token(
        self,
        uidb64: str,
        token: str,
        fail_silently: bool | None = None,
    ) -> tuple[object | None, Log | None]:
        if fail_silently is None:
            fail_silently = self.fail_silently_on_check

        # decode uidb64
        try:
            decoded_attr = decode(uidb64)
        except DjangoUnicodeDecodeError:
            return None, None

        # user lookup
        user = self.user_model.objects.filter(
            **{self.encoding_field: decoded_attr}
        ).first()
        if user is None:
            return None, None

        # check token
        checked, log = self._token_generator.check_token(
            user, token, fail_silently=fail_silently
        )
        if not checked:
            return None, None

        return user, log

    def run_callbacks(
        self,
        user: object,
        callback_kwargs: Iterable[dict[str, Any]] | None = None,
        fail_silently: bool | None = None,
    ) -> dict[str, list[Any]]:
        if fail_silently is None:
            fail_silently = self.fail_silently_on_callbacks

        callbacks_returns = self._token_generator.run_callbacks(
            user, callback_kwargs=callback_kwargs, fail_silently=fail_silently
        )

        return callbacks_returns


default_tokenizer = URLTokenizer()



from rest_framework import serializers
from rest_framework.exceptions import AuthenticationFailed

from django.contrib.auth import get_user_model
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from .enums import Channel
from .exceptions import URLTokenizerError
from .models import Log
from .tokenizer import URLTokenizer
from .utils import SETTINGS, _from_config

User = get_user_model()


EMAIL_FIELD = _from_config(SETTINGS, "email_field", "email")
PHONE_FIELD = _from_config(SETTINGS, "phone_field", "phone")


class ChannelSerializer(serializers.Serializer):
    channel = serializers.ChoiceField(choices=Channel.choices, required=True)


class SendTokenSerializer(ChannelSerializer):
    email = serializers.EmailField(required=False)
    phone = serializers.CharField(required=False)
    log = serializers.PrimaryKeyRelatedField(read_only=True)

    def validate(self, data):
        email, phone = data.get("email"), data.get("phone")
        if not email and not phone:
            raise serializers.ValidationError(
                _("Either 'email' or 'phone' is required for sending token.")
            )

        # user lookup

        self.context["user"] = get_object_or_404(
            User, **{EMAIL_FIELD: email} if email else {PHONE_FIELD: phone}
        )

        return data

    def create(self, validated_data):
        view = self.context["view"]

        token_type_url_kwarg = getattr(view, "token_type_url_kwarg", "type")
        assert token_type_url_kwarg in view.kwargs, _(
            "Expected view %s to be called with a URL keyword argument "
            "named 'type'. Fix your URL conf, or set the `.token_type_url_kwarg` "
            "attribute on the view correctly." % view.__class__.__name__
        )

        tokenizer = URLTokenizer(view.kwargs[token_type_url_kwarg])

        channel = validated_data.get("channel")

        # send token

        url_token = tokenizer.generate_tokenized_link(
            self.context.get("user"),
            path=self.context.get("path"),
            domain=self.context.get("domain"),
            protocol=self.context.get("protocol"),
            port=self.context.get("port"),
            channel=channel,
            template=self.context.get("template"),
            email_subject=self.context.get("email_subject"),
            fail_silently=self.context.get("fail_silently"),
        )

        validated_data["log"] = url_token.log

        if url_token.precondition_failed:
            raise AuthenticationFailed(_("precondition failed"))

        if url_token.exception:
            raise serializers.ValidationError(url_token.exception)

        return validated_data


class BulkSendTokenSerializer(ChannelSerializer):
    emails = serializers.ListField(
        child=serializers.EmailField(), required=False, write_only=True
    )
    phones = serializers.ListField(
        child=serializers.CharField(), required=False, write_only=True
    )
    sent = serializers.JSONField(read_only=True, default=dict)
    precondition_failed = serializers.JSONField(read_only=True, default=dict)
    errors = serializers.JSONField(read_only=True, default=dict)

    def create(self, validated_data):
        view = self.context["view"]

        token_type_url_kwarg = getattr(view, "token_type_url_kwarg", "type")
        assert token_type_url_kwarg in view.kwargs, _(
            "Expected view %s to be called with a URL keyword argument "
            "named 'type'. Fix your URL conf, or set the `.token_type_url_kwarg` "
            "attribute on the view correctly." % view.__class__.__name__
        )

        tokenizer = URLTokenizer(view.kwargs[token_type_url_kwarg])

        channel = validated_data.get("channel")

        # users lookup

        users = self.context.get("users")
        emails, phones = validated_data.get("emails"), validated_data.get("phones")
        users = users or User.objects.filter(
            Q(**{f"{EMAIL_FIELD}__in": emails}) | Q(**{f"{PHONE_FIELD}__in": phones})
        )

        # generate and send tokens

        url_tokens = tokenizer.bulk_generate_tokenized_link(
            self.context.get("users"),
            path=self.context.get("path"),
            domain=self.context.get("domain"),
            protocol=self.context.get("protocol"),
            port=self.context.get("port"),
            channel=channel,
            template=self.context.get("template"),
            email_subject=self.context.get("email_subject"),
            fail_silently=self.context.get("fail_silently"),
        )

        for url_token in url_tokens:
            to = url_token.email if channel == Channel.EMAIL else url_token.phone
            log = url_token.log.pk if url_token.log else None
            if url_token.exception:
                validated_data.setdefault("errors", {})[to] = log
            elif url_token.precondition_failed:
                validated_data.setdefault("precondition_failed", {})[to] = log
            else:
                validated_data.setdefault("sent", {})[to] = log

        return validated_data


class CheckTokenSerializer(serializers.Serializer):
    uidb64 = serializers.CharField(required=True)
    token = serializers.CharField(required=True)
    callbacks_kwargs = serializers.JSONField(required=False, write_only=True)
    callbacks_returns = serializers.JSONField(read_only=True)
    user = serializers.PrimaryKeyRelatedField(read_only=True)
    log = serializers.PrimaryKeyRelatedField(read_only=True)

    def create(self, validated_data):
        view = self.context["view"]

        token_type_url_kwarg = getattr(view, "token_type_url_kwarg", "type")
        assert token_type_url_kwarg in view.kwargs, _(
            "Expected view %s to be called with a URL keyword argument "
            "named 'type'. Fix your URL conf, or set the `.token_type_url_kwarg` "
            "attribute on the view correctly." % view.__class__.__name__
        )

        tokenizer = URLTokenizer(view.kwargs[token_type_url_kwarg])
        fail_silently = self.context.get("fail_silently")

        # check token

        uidb64, token = validated_data["uidb64"], validated_data["token"]
        user, log = tokenizer.check_token(
            uidb64, token, fail_silently=fail_silently
        )
        if user is None:
            raise AuthenticationFailed(_("The token is invalid or has expired."))

        # run callbacks

        callback_kwargs = validated_data.get("callbacks_kwargs", {})

        try:
            validated_data["callbacks_returns"] = tokenizer.run_callbacks(
                user, callback_kwargs=callback_kwargs, fail_silently=fail_silently
            )
        except URLTokenizerError as e:
            raise serializers.ValidationError(e)

        validated_data["user"] = user
        validated_data["log"] = log

        return validated_data


class LogSerializer(serializers.ModelSerializer):
    class Meta:
        model = Log
        fields = "__all__"



from collections.abc import Iterable
from enum import Enum
from typing import Any, Callable

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from .enums import Channel
from .models import Log
from .tokenizer import URLToken
from .utils import Template, encode


class URLTokenizerMixin:
    @property
    def last_channel(self) -> Channel | None:
        log = self.log_set.last()
        return log.channel if log else None

    def __init__(self, *args, **kwargs):
        # Check if the model is the auth user model
        if self.__class__ != get_user_model():
            raise ImproperlyConfigured(
                _("URLTokenizerMixin must be used with the auth user model")
            )
        super().__init__(*args, **kwargs)

    def generate_tokenized_link(
        self,
        token_type: str | Enum | None = None,
        path: str | Callable[[object], str] | None = None,
        domain: str | None = None,
        protocol: str | None = None,
        port: str | None = None,
        channel: Channel | None = None,
        template: Template | Callable[[URLToken], Template] | None = None,
        email_subject: str | None = None,
        fail_silently: bool | None = None,
    ) -> URLToken:
        from .tokenizer import URLTokenizer

        tokenizer = URLTokenizer(token_type)

        return tokenizer.generate_tokenized_link(
            self,
            path=path,
            domain=domain,
            protocol=protocol,
            port=port,
            channel=channel,
            template=template,
            email_subject=email_subject,
            fail_silently=fail_silently,
        )

    def check_token(
        self,
        token_type: str | Enum | None = None,
        token: str | None = None,
        callback_kwargs: Iterable[dict[str, Any]] | None = None,
        fail_silently: bool | None = None,
    ) -> tuple[bool, Log | None, dict[str, list[Any]]]:
        from .tokenizer import URLTokenizer

        tokenizer = URLTokenizer(token_type)
        uidb64 = encode(getattr(self, tokenizer.encoding_field))

        user, log = tokenizer.check_token(uidb64, token, fail_silently=fail_silently)
        if user is None:
            return False, log, {}

        callbacks_returns = tokenizer.run_callbacks(
            self, callback_kwargs=callback_kwargs, fail_silently=fail_silently
        )

        return True, log, callbacks_returns



from enum import Enum, unique
from typing import Any

from django.utils.translation import gettext_lazy as _


@unique
class ErrorCode(Enum):
    invalid_token_type = _("There is no token type with name '{token_type}'")
    invalid_method = _("method '{method_name}' does not exist or is not callable")
    no_email = _("User does not have an email address associated with their account")
    no_phone = _("User does not have a phone number associated with their account")
    send_precondition_execution_error = _(
        "Error during send precondition '{pred}' execution"
    )
    check_precondition_execution_error = _(
        "Error during check precondition '{pred}' execution"
    )
    callback_configuration_error = _(
        "Callback must include one of following keys: 'method', 'path', 'lambda' or 'builtin'"
    )
    invalid_builtin_callback = _(
        "Builtin callback '{builtin_name}' does not exist"
    )
    callback_execution_error = _("Error during callback '{callback}' execution")


class URLTokenizerError(Exception):
    def __init__(
        self,
        error_code: ErrorCode,
        context: dict[str, Any] | None = None,
        *args,
        **kwargs,
    ):
        self.message = error_code.value.format(*args, **kwargs)
        self.code = error_code.name
        self.context = context or {}

    def __repr__(self) -> str:
        return "URLTokenizerError({}, code={}, context={})".format(
            self.message, self.code, self.context
        )

    def __str__(self) -> str:
        return f"{self.message} (code={self.code}, context={self.context})"



import hashlib
import inspect
from collections import defaultdict
from collections.abc import Iterable
from datetime import datetime
from functools import partial, wraps
from typing import Any

from django.conf import settings
from django.db.utils import ProgrammingError
from django.utils.crypto import constant_time_compare, salted_hmac
from django.utils.encoding import force_bytes
from django.utils.http import base36_to_int, int_to_base36
from django.utils.module_loading import import_string

from .builtin_callbacks import BUILTIN_CALLBACKS
from .exceptions import ErrorCode, URLTokenizerError
from .models import Log
from .utils import _from_config, _parse_preconditions, encode


class TokenGenerator:
    """
    Strategy object used to generate and check tokens.
    """

    key_salt = "django.contrib.auth.tokens.PasswordResetTokenGenerator"
    algorithm = None
    _secret = None

    @property
    def _now(self) -> datetime:
        return datetime.now()

    @property
    def secret(self) -> str:
        return self._secret or settings.SECRET_KEY

    @secret.setter
    def secret(self, value: str):
        self._secret = value

    def __init__(self, token_config: dict[str, Any] | None = None):
        token_config = token_config or {}

        # token
        self.algorithm = self.algorithm or "sha256"
        self.encoding_field = _from_config(token_config, "encoding_field", "pk")
        self.attributes = _from_config(token_config, "attributes", [])
        self.timeout = _from_config(token_config, "timeout", 60)

        # check
        self.check_preconditions = _parse_preconditions(
            token_config, "check_preconditions"
        )
        self.check_logs = _from_config(token_config, "check_logs", False)
        self.callbacks = _from_config(token_config, "callbacks", [])

    @staticmethod
    def __num_ms(dt) -> int:
        return int((dt - datetime(2001, 1, 1)).total_seconds() * 1000)

    def _make_hash_value(self, user: object, timestamp: int) -> str:
        """
        Hash the user's primary key and some user attributes to make sure that
        the token is invalidated when the user changes these attributes.
        """
        attributes = [getattr(user, attribute) for attribute in self.attributes]
        return f"{user.pk}{timestamp}{attributes}"

    def _make_token_with_timestamp(self, user: object, timestamp: int) -> str:
        # timestamp is number of seconds since 2001-1-1. Converted to base 36,
        # this gives us a 6 digit string until about 2069.
        ts_b36 = int_to_base36(timestamp)
        hash_string = salted_hmac(
            self.key_salt,
            self._make_hash_value(user, timestamp),
            secret=self.secret,
            algorithm=self.algorithm,
        ).hexdigest()[
            ::2
        ]  # Limit to shorten the URL.
        return f"{ts_b36}-{hash_string}"

    # helpers

    def _get_log(self, user: object, token: str) -> Log | None:
        uidb64 = encode(getattr(user, self.encoding_field))
        token_hash = hashlib.sha256(force_bytes(uidb64 + token)).hexdigest()

        try:
            log = Log.objects.filter(hash=token_hash).last()
        except ProgrammingError:
            return None

        return log

    def _validate_preconditions(
        self, user: object, token: str, fail_silently: bool = False
    ) -> bool:
        for k, pred in self.check_preconditions.items():
            try:
                if not pred(user):
                    log = self._get_log(user, token)
                    if log is not None:
                        log.check_precondition_failed = k
                        log.save(update_fields=["check_precondition_failed"])
                    return False
            except Exception as e:
                if fail_silently:
                    return False

                raise URLTokenizerError(
                    ErrorCode.check_precondition_execution_error,
                    context={"exception": e},
                    pred=pred,
                ) from e

        return True

    def make_token(self, user: object) -> tuple[str, datetime]:
        """
        Return a token that can be used once for the given user.
        """
        now = self._now
        return self._make_token_with_timestamp(user, self.__num_ms(now)), now

    def check_token(
        self,
        user: object,
        token: str,
        fail_silently: bool = False,
    ) -> tuple[bool, Log | None]:
        """
        Check that a token is correct for a given user.
        """
        if not user:
            return False, None

        # Parse the token
        try:
            ts_b36, _ = token.split("-")
            ts = base36_to_int(ts_b36)
        except ValueError:
            return False, None

        # Check that the timestamp/uid has not been tampered with
        if not constant_time_compare(self._make_token_with_timestamp(user, ts), token):
            return False, None

        # Check the timestamp is within limit.
        if self.timeout and (self.__num_ms(self._now) - ts) / 1000 > self.timeout:
            return False, None

        # Check the preconditions
        if not self._validate_preconditions(user, token, fail_silently):
            return False, None

        # Check log
        log = None
        if self.check_logs:
            log = self._get_log(user, token)
            if log is None or log.checked:
                return False, None

            log._check()

        return True, log

    def run_callbacks(
        self,
        user: object,
        callback_kwargs: Iterable[dict[str, Any]] | None = None,
        fail_silently: bool = False,
    ) -> dict[str, list[Any]]:
        """
        Run the configured callbacks for the given user. Callbacks can be
        configured to receive kwargs, and to return values that will be collected
        and returned by this method.
        """

        def resolve_callback(callback: dict[str, Any]):
            if callback.get("method"):
                return getattr(user, callback["method"], None)

            if callback.get("path"):
                return import_string(callback["path"])

            if callback.get("lambda"):
                return partial(callback["lambda"], user)

            if callback.get("builtin"):
                builtin_name = callback["builtin"]
                func = BUILTIN_CALLBACKS.get(builtin_name)
                if func is None:
                    raise URLTokenizerError(
                        ErrorCode.invalid_builtin_callback,
                        builtin_name=builtin_name,
                    )

                @wraps(func)
                def builtin_wrapper(**kwargs):
                    return func(user, **kwargs)

                return builtin_wrapper

            raise URLTokenizerError(ErrorCode.callback_configuration_error)

        def extract_kwargs(
            method, available_kwargs: list[dict[str, Any]], defaults: dict
        ):
            signature = inspect.signature(method)
            param_names = set(signature.parameters.keys())

            for i, kw in enumerate(available_kwargs):
                if set(kw.keys()).issubset(param_names):
                    selected = available_kwargs.pop(i)
                    break
            else:
                selected = {}

            final_kwargs = {**selected, **defaults}

            return final_kwargs

        available_kwargs = list(callback_kwargs or [])
        results = defaultdict(list)

        for callback in self.callbacks:
            try:
                method = resolve_callback(callback)

                if not callable(method):
                    raise URLTokenizerError(
                        ErrorCode.invalid_method,
                        method_name=method.__name__ if method else None,
                    )

                kwargs = extract_kwargs(
                    method,
                    available_kwargs,
                    callback.get("defaults", {}),
                )

                result = method(**kwargs)

                if callback.get("return_value"):
                    results[method.__name__].append(result)

            except Exception as e:
                if not fail_silently:
                    raise URLTokenizerError(
                        ErrorCode.callback_execution_error,
                        context={"exception": e},
                        callback=method.__name__ if method else None,
                    ) from e

        return results


# A singleton instance to use by default
# expiration time is 60 seconds by default
default_token_generator = TokenGenerator()



from typing import Any

from django.utils.module_loading import import_string

from .utils import SETTINGS


def serialize_user(
    user: object,
    *,
    related_serializers: dict[str, str] | None = None,
    **kwargs,
) -> dict[str, Any] | None:
    """
    Built-in callback that serializes the user using the configured USER_SERIALIZER.

    - USER_SERIALIZER must be a string path to a DRF serializer class.
    - related_serializers is an optional dict where:
        key   -> field_name on user
        value -> string path to DRF serializer class
    """

    user_serializer = SETTINGS.get("USER_SERIALIZER")
    if user_serializer is None:
        return None

    serializer_class = import_string(user_serializer)
    data = serializer_class(user).data

    if not related_serializers:
        return data

    for field_name, serializer_path in related_serializers.items():
        # Skip if user doesn't have the attribute
        if not hasattr(user, field_name):
            continue

        related_instance = getattr(user, field_name, None)
        if related_instance is None:
            continue

        serializer_cls = import_string(serializer_path)

        serialized_related = serializer_cls(related_instance).data

        data[field_name] = serialized_related

    return data


def patch_user(
    user: object, *, data: dict[str, Any] = None, **kwargs
) -> dict[str, Any] | None:
    """
    Built-in callback that updates the user using the configured USER_SERIALIZER.

        The USER_SERIALIZER setting should be a string path to a DRF serializer class that
        takes the user as input and updates it with the provided data. The updated user
        data is returned as a dictionary to be used in the callback.
    """

    user_serializer = SETTINGS.get("USER_SERIALIZER")
    if user_serializer is None or data is None:
        return None

    serializer_class = import_string(user_serializer)

    serializer = serializer_class(user, data=data, partial=True)
    serializer.is_valid(raise_exception=True)
    serializer.save()

    return serializer.data


# Registry mapping built-in callback names to their implementations.
# Add new built-in callbacks here to make them available via the
# ``builtin`` key in the callback configuration.
BUILTIN_CALLBACKS = {
    "serialize_user": serialize_user,
    "patch_user": patch_user,
}



from enum import Enum
from typing import Callable

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

from .enums import Channel
from .tokenizer import URLToken
from .utils import Template


class URLTokenizerQueryset(models.QuerySet):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Check if the QuerySet's model is the auth user model
        if self.model and self.model != get_user_model():
            raise ImproperlyConfigured(
                _("URLTokenizerManager must be used with the auth user model")
            )

    def bulk_generate_tokenized_link(
        self,
        token_type: str | Enum | None = None,
        path: str | Callable[[object], str] | None = None,
        domain: str | None = None,
        protocol: str | None = None,
        port: str | None = None,
        channel: Channel | None = None,
        template: Template | Callable[[URLToken], Template] | None = None,
        email_subject: str | None = None,
        fail_silently: bool | None = None,
    ) -> list[URLToken]:
        from .tokenizer import URLTokenizer

        tokenizer = URLTokenizer(token_type)

        return tokenizer.bulk_generate_tokenized_link(
            self,
            path=path,
            domain=domain,
            protocol=protocol,
            port=port,
            channel=channel,
            template=template,
            email_subject=email_subject,
            fail_silently=fail_silently,
        )


URLTokenizerManager = models.Manager.from_queryset(URLTokenizerQueryset)
